/*
 * usbhw.c : LPC2148 USB logic driver
 *   by uratan! 2008.9.22
 */
#include "common.h"
#include "LPC214x.h"
#include "irq.h"
#include "dscr.h.S"
#include "usbhw.h"

/*
 * my local debug flag
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf

/*
 * register bitmaps
 */
/* DEV_INT_EN:   Device Interrupt Enable Register */
/* DEV_INT_STAT: Device Interrupt Status Register */
/* DEV_INT_CLR:  Device Interrupt Clear  Register */
#define UDI_FRAME_INT       0x00000001
#define UDI_EP_FAST_INT     0x00000002
#define UDI_EP_SLOW_INT     0x00000004
#define UDI_DEV_STAT_INT    0x00000008
#define UDI_CCEMTY_INT      0x00000010
#define UDI_CDFULL_INT      0x00000020
#define UDI_RxENDPKT_INT    0x00000040
#define UDI_TxENDPKT_INT    0x00000080
#define UDI_EP_RLZED_INT    0x00000100
#define UDI_ERR_INT         0x00000200
/* RX_PLENGTH: Receive Packet Length Register */
#define PKT_LENGTH_MASK     0x000003FF
#define PKT_DV              0x00000400
#define PKT_RDY             0x00000800
/* USB_CTRL: USB Control Register */
#define UCTRL_RD_EN         0x00000001
#define UCTRL_WR_EN         0x00000002
/* CMD_CODE: USB Command Code Register */
#define CMD_SET_ADDR        0x00D00500
#define CMD_CFG_DEV         0x00D80500
//#define CMD_SET_MODE        0x00F30500
#define CMD_RD_FRAME        0x00F50500
#define DAT_RD_FRAME        0x00F50200
//#define CMD_RD_TEST         0x00FD0500
//#define DAT_RD_TEST         0x00FD0200
#define CMD_SET_DEV_STAT    0x00FE0500
#define CMD_GET_DEV_STAT    0x00FE0500
#define DAT_GET_DEV_STAT    0x00FE0200
//#define CMD_GET_ERR_CODE    0x00FF0500
//#define DAT_GET_ERR_CODE    0x00FF0200
#define CMD_RD_ERR_STAT     0x00FB0500
#define DAT_RD_ERR_STAT     0x00FB0200
#define DAT_WR_BYTE(x)     (0x00000100 | ((x) << 16))
#define CMD_SEL_EP(x)      (0x00000500 | ((x) << 16))
#define DAT_SEL_EP(x)      (0x00000200 | ((x) << 16))
//#define CMD_SEL_EP_CLRI(x) (0x00400500 | ((x) << 16))
//#define DAT_SEL_EP_CLRI(x) (0x00400200 | ((x) << 16))
#define CMD_SET_EP_STAT(x) (0x00400500 | ((x) << 16))
#define CMD_CLR_BUF         0x00F20500
//#define DAT_CLR_BUF         0x00F20200
#define CMD_VALID_BUF       0x00FA0500
/* Device Status Register */
#define DSTAT_CONN          0x01
#define DSTAT_CONN_CH       0x02
#define DSTAT_SUSP          0x04
#define DSTAT_SUSP_CH       0x08
#define DSTAT_RST           0x10

/*
 * functions
 */
int uhw_Init(void);
void uhw_Connect(int conn);
void uhw_WakeUp(void);
void uhw_SetAddress(u_char addr);
void uhw_Configure(int cfg);
void uhw_EnableEP(u_char EPaddr, u_long max_packet_size);
u_long uhw_IsEnabledEP(u_char EPaddr);
u_long uhw_IsStallEP(u_char EPaddr);
void uhw_SetStallEP(u_char EPaddr);
void uhw_ClrStallEP(u_char EPaddr);
void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len);
void uhw_arm_WriteEP(u_char EPaddr);
u_long uhw_ReadEP(u_char EPaddr, u_char *pData, int capacity);
void uhw_ReadEP_finish(u_char EPaddr);
static void ucmd_with_data(u_long cmd, u_long cmd2);
static u_long ucmd_to_read_data(u_long cmd);
static void ucmd_single_command(u_long cmd);
static u_long get_EPnum(u_long EPaddr);
static void USB_ISR(void) __attribute__ ((interrupt("IRQ")));

	/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * 'EPaddr' is USB spec style Endpoint address.
	 *   Endpoint number is at [bit3..0], bit 7 has direction
	 * 'EPnum' is LPC2148 physical Endpoint number.
	 *   Endpoint number is at [bit4..1], bit 0 has direction
	 *vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv */

/*
 * externs
 */
/* in usbcore.c */
extern void ucore_ResetCore(void);
extern void ucore_SuspendCore(void);
extern void ucore_isr_EP0_in(u_long unused);
extern void ucore_isr_EP0_out(u_long unused);
/* in periph.c */
extern const INTR_subFUNCs USB_Jump_Table[USB_N_JT];
extern const INTR_subFUNCs USB_Jump_Table_EP[USB_N_EP];

/*
 * variables
 */
/* in main.c */
extern VOLATILE int GotSUD;		/* SetUp Data */
extern VOLATILE int GotCOD;		/* Control-OUT Data */
extern VOLATILE int Sleep;		/* shall Sleep becaue in Suspend mode */


/* =============================================================== */

/*
 * initialize USB hardware
 */
int uhw_Init(void)
{
	/*
	 * Turn on USB logic
	 */
	PCONP |= PP_PUSB;			/* Turn On USB PCLK */

	/*
	 * Enable Vbus sense (P0.23)
	 */
	PINSEL1 |= (1<<14);

	/*
	 * select P0.31 pin function
	 *   1 - GoodLinc
	 *   2 - ~Connect
	 */
	PINSEL1 |= (2<<30);

	/*
	 * Configure 48MHz USB Clock
	 */
	cm_usb_pll_init();

	/*
	 * setup vectored interrupt
	 */
	if(install_irq(USB_INT, (void *)USB_ISR) == FALSE) {
		return(FALSE);
	}

	/*
	 * setup hardware buffer for Control Endpoint
	 *   it takes about 6usec (24MHz+24MHz)
	 */
	DEV_INT_CLR = UDI_EP_RLZED_INT;
	REALIZE_EP = 0x00000003;		/* reset value, not needed */
	EP_INDEX = 0;
	MAXPACKET_SIZE = DSCR_MPSIZE_EP0;
	EP_INDEX = 1;
	MAXPACKET_SIZE = DSCR_MPSIZE_EP0;
	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;
		/*
		 * MAX_PACKET_SIZEs of each Endpoint are not reset by USB-RESET,
		 * it seems that only REALIZE_EP register may be reset by USB-RESET.
		 */

	/*
	 * Enable USB interrupt
	 */
	EP_INT_PRIO = 0x00000000;
	EP_INT_CLR  = 0xFFFFFFFF;
	EP_INT_EN   = 0xFFFFFFFF;		/* <== see note1 */
	DEV_INT_PRIO = 0x00000000;
	DEV_INT_CLR  = 0xFFFFFFFF;
	DEV_INT_EN   = UDI_DEV_STAT_INT |
					((USB_Jump_Table[USB_JT_SOF] != NULL) ? UDI_FRAME_INT : 0) |
					((USB_Jump_Table[USB_JT_ERROR] != NULL) ? UDI_ERR_INT : 0) |
					UDI_EP_SLOW_INT;
	USB_INT_STAT |= (1<<31);
		/*
		 * note1:
		 *   EP_INT_EN register will NOT mask interrupt requests !!
		 *   only switch 'slave-transfer-mode' and 'normal-mode'
		 * note2:
		 *   EP_INT_CLR will generate reponse on CMD_DATA register !
		 * note3:
		 *   to clear UDI_EP_SLOW_INT, the condition 'EP_INT_STAT==0' needed.
		 *   and UDI_EP_SLOW_INT is NOT cleared by clearing EP_INT_STAT.
		 *     (you shall clear EP_INT_STAT first, then UDI_EP_SLOW_INT)
		 */

	return(TRUE);
}

/*
 * connect USB device to the host
 *   pass 1 to Connect, 0 to Disconnect
 */
void uhw_Connect(int conn)
{
	u_long imf;

	imf = IntDisable();

	ucmd_with_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(conn ? DSTAT_CONN : 0));

	IntRestore(imf);
}

/*
 * issue USB-Remote-Wakeup action to wake up the host
 *   this is identical to uhw_Connect(1)
 */
void uhw_WakeUp(void)
{
	u_long imf;

	imf = IntDisable();

	ucmd_with_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(DSTAT_CONN));
		/*
		 * I want to write 0 to DSTAT_SUSP bit to do 'remote-wakeup' operation.
		 * if DSTAT_SUSP bit is already 0, this is ignored.
		 */

	IntRestore(imf);
}

/*
 * set device address
 */
void uhw_SetAddress(u_char addr)
{
	u_long imf;

	imf = IntDisable();

	ucmd_with_data(CMD_SET_ADDR, DAT_WR_BYTE(0x80 | addr));
		/*
		 * new addr is valid after Status-IN received by hardware
		 */

	IntRestore(imf);
}

/*
 * Configure the USB device
 *   pass 1 to Configure, 0 to De-configure
 *   (in both case, non-control-Endpoints are disabled)
 */
void uhw_Configure(int cfg)
{
	u_long imf;

	imf = IntDisable();

	/*
	 * configure/deconfigure the device
	 */
	ucmd_with_data(CMD_CFG_DEV, DAT_WR_BYTE(cfg ? 1 : 0));

	/*
	 * disable all Endpoints except control-pipe
	 *   it takes about 4usc (24MHz+24MHz)
	 */
	DEV_INT_CLR = UDI_EP_RLZED_INT;
	REALIZE_EP = 0x00000003;
	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;

	IntRestore(imf);
}

/*
 * realize to enable the Endpoint
 */
void uhw_EnableEP(u_char EPaddr, u_long max_packet_size)
{
	u_long EPnum;
	u_long imf;

	EPnum = get_EPnum(EPaddr);

	imf = IntDisable();

	/*
	 * enable specified Endpoint with allocating buffer
	 *   it takes about 5usc (24MHz+24MHz)
	 */
	DEV_INT_CLR = UDI_EP_RLZED_INT;
	REALIZE_EP |= (1<<EPnum);
	EP_INDEX = EPnum;
	MAXPACKET_SIZE = max_packet_size;
	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;

	IntRestore(imf);
}

/*
 * check the Endpoint is enabled or not
 *   return:  1 if enabled
 */
u_long uhw_IsEnabledEP(u_char EPaddr)
{
	u_long EPnum, enabled;

	EPnum = get_EPnum(EPaddr);
	enabled = (REALIZE_EP & (1<<EPnum)) ? TRUE : FALSE;

	return(enabled);
}

/*
 * check STALL bit of the Endpoint
 *   return:  1 if stalled
 */
u_long uhw_IsStallEP(u_char EPaddr)
{
	u_long EPnum, val;
	u_long imf;

	EPnum = get_EPnum(EPaddr);

	imf = IntDisable();

	/*
	 * examine 'ST' bit of Select-Endpoint-status value
	 */
	ucmd_single_command(CMD_SEL_EP(EPnum));
	val = (ucmd_to_read_data(DAT_SEL_EP(EPnum)) & /*ST*/ 0x02) ? 1 : 0;

	IntRestore(imf);

	return(val);
}

	/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * 'ST' bit is at: (1<<1) in Select-Endpoint-status value.
	 *                 (1<<0) in Set-Endpoint-status register.
	 *vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv */

/*
 * set STALL bit of the Endpoint
 */
void uhw_SetStallEP(u_char EPaddr)
{
	u_long imf;

	imf = IntDisable();

	/*
	 * set 'ST' bit of Set-Endpoint-status register
	 */
	ucmd_with_data(CMD_SET_EP_STAT(get_EPnum(EPaddr)), DAT_WR_BYTE(/*ST*/ 0x01));

	IntRestore(imf);
}

/*
 * clear STALL bit of the Endpoint
 */
void uhw_ClrStallEP(u_char EPaddr)
{
	u_long imf;

	imf = IntDisable();

	ucmd_with_data(CMD_SET_EP_STAT(get_EPnum(EPaddr)), DAT_WR_BYTE(0));

	IntRestore(imf);
}

/*
 * write data to IN Endpoint buffer
 */
void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len)
{
	u_long imf;
	u_long i;

	imf = IntDisable();

	/*
	 * open the path
	 */
	USB_CTRL = ((EPaddr & 0x0F) << 2) | UCTRL_WR_EN;

	/*
	 * set data length, then write
	 */
	TX_PLENGTH = len;

	len = (len + 3) / 4;
	for(i=0; i<len; i++) {
		TX_DATA = pData[0] + (pData[1]<<8) + (pData[2]<<16) + (pData[3]<<24);
		pData += 4;
	}

	/*
	 * close the path
	 */
	USB_CTRL = 0;

	IntRestore(imf);
}

/*
 * permit hardware buffer to send the data
 */
void uhw_arm_WriteEP(u_char EPaddr)
{
	u_long imf;

	imf = IntDisable();

	ucmd_single_command(CMD_SEL_EP(get_EPnum(EPaddr)));
	ucmd_single_command(CMD_VALID_BUF);

	IntRestore(imf);
}

/*
 * read data from OUT Endpoint buffer
 *   pass your buffer capacity to avoid buffer-overflow
 *   return: number of bytes read
 */
u_long uhw_ReadEP(u_char EPaddr, u_char *pData, int capacity)
{
	u_long imf;
	u_long len, i, rx;

	imf = IntDisable();

	/*
	 * open the path
	 */
	USB_CTRL = ((EPaddr & 0x0F) << 2) | UCTRL_RD_EN;

	/*
	 * get received data length
	 */
	do {
		len = RX_PLENGTH;
	} while((len & PKT_RDY) == 0);
	len &= PKT_LENGTH_MASK;

	/*
	 * read the data
	 *   stop reading if overflow the buffer
	 */
	for(i=0; i<len; i++) {
		if(--capacity < 0)
			break;

		if((i % 4) == 0) {
			rx = RX_DATA;
		}
		*pData++ = rx & 0xff;
		rx >>= 8;
	}

	/*
	 * close the path
	 */
	USB_CTRL = 0;

	IntRestore(imf);

	return(len);
}

/*
 * permit hardware buffer to receive next OUT
 */
void uhw_ReadEP_finish(u_char EPaddr)
{
	u_long imf;

	imf = IntDisable();

	ucmd_single_command(CMD_SEL_EP(get_EPnum(EPaddr)));
	ucmd_single_command(CMD_CLR_BUF);

	IntRestore(imf);
}

/*
 * protocol engine: write a command with data
 *   it takes about 4usec (24MHz+24MHz)
 */
static void ucmd_with_data(u_long cmd, u_long cmd2)
{
	/*
	 * first command (should be CMD_XXX)
	 */
	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;

	/*
	 * second command (should be DAT_XXX)
	 */
	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = cmd2;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;
}

/*
 * protocol engine: write a command to get data
 *   it takes about 2.6usec (24MHz+24MHz)
 */
static u_long ucmd_to_read_data(u_long cmd)
{
	/*
	 * apply command (should be DAT_XXX)
	 */
	DEV_INT_CLR = UDI_CCEMTY_INT | UDI_CDFULL_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;

	/*
	 * wait for the result ready
	 */
	while((DEV_INT_STAT & UDI_CDFULL_INT) == 0)
		;

	return(CMD_DATA);
}

/*
 * protocol engine: write a command
 *   it takes about 2usec (24MHz+24MHz)
 */
static void ucmd_single_command(u_long cmd)
{
	/*
	 * apply command (should be CMD_XXX)
	 */
	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;
}

/*
 * convert 'bEndpointAddress' to LPC2148 physical Endpoint number
 *   example: 0x00(ep0out) ==> 0x000001
 *            0x80(ep0in)  ==> 0x000002
 *            0x01(ep1out) ==> 0x000003
 *            0x81(ep1in)  ==> 0x000004
 */
static u_long get_EPnum(u_long EPaddr)
{
	u_long EPnum;

	EPnum = (EPaddr & 0x0F) << 1;
	if(EPaddr & 0x80) {
		EPnum += 1;
	}
	return(EPnum);
}

/* =============================================================== */

/*
 * USB Interrupt Service Routine
 */
static void USB_ISR(void)
{
	u_long dis, eis;
	u_long val, i, n;

SET_DEBUG_PORT2;

	dis = DEV_INT_STAT;				/* Device Interrupt Status */
RPRINTF("DEV_INT_STAT:%04x\r\n", dis);
	dis &= DEV_INT_EN;

	if(dis & UDI_DEV_STAT_INT) {
		/*
		 * Device Status Interrupt (Reset, Suspend/Resume, Connect change)
		 *   get device status detail
		 */
		ucmd_single_command(CMD_GET_DEV_STAT);
		val = ucmd_to_read_data(DAT_GET_DEV_STAT);
RPRINTF(" val:%04x\r\n", val);

		if(val & DSTAT_RST) {
			/*
			 * USB Reset
			 *   force clear all other interrupt requests
			 */
			EP_INT_CLR  = 0xFFFFFFFF;
			DEV_INT_CLR = 0xFFFFFFFF;

			/*
			 * reset params on other layer
			 */
			GotSUD = FALSE;
			GotCOD = FALSE;
			Sleep = FALSE;
			ucore_ResetCore();
			if(USB_Jump_Table[USB_JT_RESET] != NULL) {
				USB_Jump_Table[USB_JT_RESET](0);
			}
			goto isr_end;
		}

		if(val & DSTAT_SUSP_CH) {
			/*
			 * Suspend/Resume
			 *   these event is generated from 'USB_need_clock',
			 *   in other word, from USB bus activities.
			 */
			if(val & DSTAT_SUSP) {
				/*
				 * Suspend
				 *   interrupt is generated after confirmation
				 *   by 2sec-delay-filter.
				 */
				GotSUD = FALSE;
				GotCOD = FALSE;
				Sleep = TRUE;
				ucore_SuspendCore();
				if(USB_Jump_Table[USB_JT_SUSPEND] != NULL) {
					USB_Jump_Table[USB_JT_SUSPEND](0);
				}
			} else {
				/*
				 * Resume
				 *   interrupt is immediately generated.
				 *   Resume interrupt will also come when power-on because
				 *   power-on default value of 'USB_need_clock' may be '1'.
				 *     (normaly with Reset interrupt, and ignored by this code)
				 */
//				GotSUD = FALSE;
//				GotCOD = FALSE;
//				Sleep = FALSE;
				if(USB_Jump_Table[USB_JT_RESUME] != NULL) {
					USB_Jump_Table[USB_JT_RESUME](0);
				}
			}

			goto isr_end;
		}

		if(val & DSTAT_CONN_CH) {
			/*
			 * Connect change (Vbus voltage appeared/disappeared)
			 */
//			GotSUD = FALSE;
//			GotCOD = FALSE;
//			Sleep = FALSE;
			if(USB_Jump_Table[USB_JT_POWER] != NULL) {
				USB_Jump_Table[USB_JT_POWER](val & DSTAT_CONN);
			}
			goto isr_end;
		}
	}

	if(dis & UDI_FRAME_INT) {
		/*
		 * Start of Frame Interrupt
		 */
		if(USB_Jump_Table[USB_JT_SOF] != NULL) {
			ucmd_single_command(CMD_RD_FRAME);
			val = ucmd_to_read_data(DAT_RD_FRAME);
			val = val | (ucmd_to_read_data(DAT_RD_FRAME) << 8);
			USB_Jump_Table[USB_JT_SOF](val);
		}
	}

	if(dis & UDI_ERR_INT) {
		/*
		 * Error Interrupt
		 */
		if(USB_Jump_Table[USB_JT_ERROR] != NULL) {
			ucmd_single_command(CMD_RD_ERR_STAT);
			val = ucmd_to_read_data(DAT_RD_ERR_STAT);

			if        (val & 0x01) {
				val = USB_ERR_PID;
			} else if(val & 0x02) {
				val = 0x100;	// Unexpected Error
			} else if(val & 0x04) {
				val = USB_ERR_CRC;
			} else if(val & 0x08) {
				val = USB_ERR_TIMEOUT;
			} else if(val & 0x10) {
				val = USB_ERR_EOP;
			} else if(val & 0x20) {
				val = 0x101;	// Buffer Overrun
			} else if(val & 0x40) {
				val = USB_ERR_BIT_STUFF;
			} else if(val & 0x80) {
				val = USB_ERR_DATA_TOGGLE;
			}
			USB_Jump_Table[USB_JT_ERROR](val);
		}
	}

	if(dis & UDI_EP_SLOW_INT) {
		/*
		 * Ednpoint's Slow Interrupt
		 */
		while((eis = EP_INT_STAT)) {
RPRINTF(" eis:%08x\r\n", eis);
			/*
			 * Check All Endpoints
			 */
	/*
	 * ABOUT HANDSHAKE WITH HOST BY SETUP REQUEST FLOW. (to debug with uart)
	 *
	 * the Endpoint interrupt is generated when:
	 *   [SETUP], [OUT] ... when the data arrived
	 *   [IN]           ... after the data transmitted
	 *
	 * USB hardware may act like these.
	 *   [SETUP] ... always accepted, if previous [OUT] data is still
	 *               in hardware buffer, the data is overwritten.
	 *   [IN]    ... host should wait until the firmware set the data.
	 *   [OUT]   ... always accepted if the hardware buffer is free.
	 *
	 * SETUP request flows like these.
	 *   1.Control-Write:
	 *         [SETUP] [IN]..[IN(*)] [OUT as stat]  --> [nextSETUP]
	 *                             ^1            ^2               ^3
	 *   2.Control-Read:
	 *         [SETUP] [OUT]..[OUT] [IN as stat(*)] --> [nextSETUP]
	 *                                            ^4              ^5
	 *   3.Control-no-data:
	 *         [SETUP] [IN as stat(*)] --> [nextSETUP]
	 *                               ^6              ^7
	 * so, if the firmware can not follow the interrupt realtimely,
	 * multiple interrupt request may be detected.
	 *   case-A) if the firmware delay catching ^1, both IN+OUT request
	 *           appeared behind ^2.
	 *   case-B) more about case-A, if the firmware delay catching ^1+^2,
	 *           both IN+OUT request appeared behind ^3, with
	 *           'SETUP-overlapped' status.
	 *   case-C) if the firmware catch ^1 but delay catching ^2, OUT request
	 *           is overwritten behind ^3, with 'SETUP-overlapped' status.
	 *   case-D) if the firmware delay catching ^4, both IN+OUT request
	 *           appeared behind ^5.
	 *   case-E) if the firmware delay catching ^6, both IN+OUT request
	 *           appeared behind ^7.
	 *
	 * in those case, we shall proceed [IN] request first before [OUT] request.
	 * (we may not have jobs about this finish-of-last-IN)
	 *
	 * at the result, after armimg [IN(*)] Endpoint, we can NOT make the
	 * host wait to proceed next.
	 *
	 * also, you shall care about synchronize polling context and interrupt
	 * handler context because I will process SETUP request in polling context.
	 */
		/*
		 * ܂ UART ŃfobObZ[WĂĂoĂŎԂ
		 * ĂŜ̓ȂȂ(ƁASETUP NGXg
		 * |[O[`ŏ邽)ɂ́A炪uɐiłv
		 * ƂȂzXg҂ĂKv̂łAꂪ
		 * ł̂͌ꂽƂɌ肳A킯łB
		 * IN ͂炪f[^pӂ܂ł̓zXg͑҂
		 * Ȃ킯łA(obt@̎)OUT Endpoint 
		 * n[hM OK ႤƃzXg͎ɍsႤ킯ŁA
		 * u݃nhvɍsĂ݂҂鎟()
		 * Cxg܂ŋNĂvꍇPAKv킯łB
		 */
			for(i=0; i<USB_N_EP; i++) {
				/*
				 * exchange priority for EP0in/EP0out
				 */
				if(i==0)
					n = 1;
				else if(i==1)
					n = 0;
				else
					n = i;

				if(eis & (1<<n)) {
//					DEV_INT_CLR = UDI_CDFULL_INT;
					EP_INT_CLR = (1<<n);
					while((DEV_INT_STAT & UDI_CDFULL_INT) == 0)
						;
					val = CMD_DATA;

					if(n == 0) {
						/* control out */
RPRINTF(" val:%02x\r\n", val);
						if(val & 0x04) {
							/* this is Setup Packet */
							GotSUD = TRUE;
							GotCOD = FALSE;
							if(USB_Jump_Table[USB_JT_SUDAV] != NULL) {
								USB_Jump_Table[USB_JT_SUDAV](0);
							}
						} else {
							ucore_isr_EP0_out(0);
						}
					} else if(n == 1) {
						/* control in */
						ucore_isr_EP0_in(0);
					} else {
						/* generic Endpoints */
						if(USB_Jump_Table_EP[n] != NULL) {
							USB_Jump_Table_EP[n](0);
						}
					}
				}
			}
		}
	}

isr_end:
	/*
	 * when jumping to here from 'detecting-Device-Status-change',
	 * ignore other interrupts if exist. (except UDI_EP_SLOW_INT)
	 */
	DEV_INT_CLR = dis;

	VICVectAddr = 0;				/* Acknowledge Interrupt */
CLR_DEBUG_PORT2;
}
